# IRC desklet for gDesklets (http://gdesklets.gnomedesktop.org)
# Copyright (C) 2003  Seth Remington

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from sensor.Sensor import Sensor
from main import admin
from twisted.internet import reactor
from utils.datatypes import *
from utils import i18n
from utils import singleton
from config.ConfigManager import ConfigManager
import os
import ircbot
from message_window import message_window

class IRCSensor(Sensor):
    def __init__(self, *args):
        self.is_initialized = 0
        self.border_width = 0
        self.border_height = 0
        global _; _ = i18n.Translator("irc_sensor")
        self.gettext = _

        Sensor.__init__(self)

        # Set up configuration
        self._set_config_type("server", TYPE_STRING, "irc.gnome.org")
        self._set_config_type("port", TYPE_STRING, "6667")
        self._set_config_type("password", TYPE_STRING, "")
        self._set_config_type("encoding", TYPE_STRING, "utf-8")
        self._set_config_type("channel", TYPE_STRING, "#gdesklets")
        self._set_config_type("nick", TYPE_STRING, "")
        self._set_config_type("username", TYPE_STRING, os.getenv("LOGNAME"))
        self._set_config_type("num_lines", TYPE_STRING, "10")
        self._set_config_type("max_chars", TYPE_STRING, "100")
        self._set_config_type("highlight_color", TYPE_STRING, "red")
        self._set_config_type("bool_time", TYPE_INT, 1)
        self._set_config_type("bool_log", TYPE_INT, 0)
        self._set_config_type("log_file", TYPE_STRING, "")
        self._set_config_type("bool_tts", TYPE_INT, 0)
        self._set_config_type("tts_command", TYPE_STRING, "festival --tts")
        self._set_config_type("msg_window_display_action", TYPE_STRING, "Mouse Over")

        self._watch_config(self.on_watch_config)

        self._add_timer(1000, self.ircbot_start)

        
    def on_watch_config(self, key, value):
        if key == "server":
            pass
        elif key == "port":
            pass
        elif key == "password":
            pass
        elif key == "encoding":
            pass
        elif key == "channel":
            pass
        elif key == "nick":
            pass
        elif key == "username":
            pass
        elif key == "num_lines":
            self.num_lines = value
        elif key == "max_chars":
            self.max_chars = value
        elif key == "highlight_color":
            self.highlight_color = value
        elif key == "bool_time":
            self.bool_time = value
        elif key == "bool_log":
            self.bool_log = value
        elif key == "log_file":
            self.log_file = value
        elif key == "bool_tts":
            self.bool_tts = value
        elif key == "tts_command":
            self.tts_command = value
        elif key == "msg_window_display_action":
            self.msg_window_display_action = value

    def call_action(self, action, path, args = []):
        if action == "show_or_hide_message_window":
            if self.msg_window_display_action == "Mouse Click":
                try:
                    geometry = self.get_desklet_location()
                    self.msg_window.show_or_hide(geometry[0] + 15, geometry[1] + geometry[3])
                except:
                    try:
                        self.msg_window.show_or_hide(15, 15)
                    except:
                        pass
        if action == "show_message_window":
            if self.msg_window_display_action == "Mouse Over":
                try:
                    geometry = self.get_desklet_location()
                    self.msg_window.show(geometry[0] + 15, geometry[1] + geometry[3])
                except:
                    try:
                        self.msg_window.show(15, 15)
                    except:
                        pass
        if action == "hide_message_window":
            if self.msg_window_display_action == "Mouse Over":
                try:
                    self.msg_window.hide()
                except:
                    pass
        if action == "on_menu":
            menu = [("Reconnect", 1, None, self.reconnect_desklet, [])]
            self._open_menu(menu)
        if action == "scroll_history":
            self.IRCFactory.protocol_instance.displayer.scroll(args[0])
        if action == "scroll_clients":
            self.IRCFactory.protocol_instance.displayer.scroll_clients(args[0])

    def get_base_id(self):
        id = self._get_config_id()
        parse_id = ""
        for c in id:
            if c.isdigit():
                parse_id += c
            else:
                break
        return parse_id

    def get_desklet_location(self):
        parse_id = self.get_base_id()
        geometry = admin.get_geometry(parse_id)
        return geometry

    def ircbot_start(self):
        self.server = self._get_config("server")
        self.port = self._get_config("port")
        self.password = self._get_config("password")
        self.encoding = self._get_config("encoding")
        self.channel = self._get_config("channel")
        self.nickname = self._get_config("nick")
        self.username = self._get_config("username")
        self.num_lines = self._get_config("num_lines")
        self.max_chars = self._get_config("max_chars")
        self.highlight_color = self._get_config("highlight_color")
        self.bool_time = self._get_config("bool_time")
        self.bool_log = self._get_config("bool_log")
        self.log_file = self._get_config("log_file")
        self.bool_tts = self._get_config("bool_tts")
        self.tts_command = self._get_config("tts_command")
        self.msg_window_display_action = self._get_config("msg_window_display_action")
        
        # Watch for changes in the display's width and height
        # so we can update the dynamic borders
        if not self.is_initialized:
            self.config_manager = singleton.get(ConfigManager)
            profile = self.config_manager.get("profile")
            base_id = self.get_base_id()
            id = base_id + "_default_"
            self.config_manager.watch("profiles", profile, id, self.watch_borders)

        if not self.server or not self.port or not self.channel \
        or not self.nickname or not self.num_lines:
            output = self._new_output()
            output.set("msg", "Please configure the desklet, then\nright click and select 'Reconnect'.")
            self._send_output(output)
            self.connector = None
            return 0

        if not self.is_initialized:
            self.msg_window = message_window(0, 0, self.send_message, self.tab_complete)
            self.IRCFactory = ircbot.IRCFactory(self)

        self.IRCFactory.change_nickname(self.nickname)
        self.IRCFactory.change_channel(self.channel)
        self.connector = reactor.connectTCP(self.server, int(self.port), self.IRCFactory)
        self.connector.irc_is_running = 1

        if not self.is_initialized:
            reactor.startRunning(installSignalHandlers=0)
            self._add_timer(250, self.ircbot_loop)
            self.is_initialized = 1
            
        return 0

    def ircbot_stop(self):
        if self.connector:
            self.IRCFactory.protocol_instance.quit()
            self.connector.irc_is_running = 0
            self.connector.disconnect()
            # We need to zero out the clients
            self.IRCFactory.displayer.reset_clients()
        
    def ircbot_loop(self):
        reactor.runUntilCurrent()
        reactor.doIteration(0.002)
        return 1

    def send_message(self, message):
        self.IRCFactory.protocol_instance.send_message(self.channel, message)

    def tab_complete(self, widget, start):
        self.IRCFactory.protocol_instance.tab_complete(widget, start)
        
    def get_configurator(self):
        configurator = self._new_configurator()
        configurator.set_name(_("IRC"))
        configurator.add_title(_("IRC Settings"))
        configurator.add_entry(_("IRC Server:"), "server", _("The IRC server to log into."))
        configurator.add_entry(_("Port:"), "port", _("The port the server is listening to."))
        configurator.add_entry(_("Password:"), "password", _("The password if the server isn't public."))
        configurator.add_entry(_("Encoding:"), "encoding", _("The encoding to use in IRC connections."))
        configurator.add_entry(_("Channel:"), "channel", _("The channel to join. Don't forget the # :-)"))
        configurator.add_entry(_("Nickname:"), "nick", _("Your channel nick"))
        configurator.add_entry(_("Username:"), "username", _("Your connection username"))
        configurator.add_title(_("Display"))
        configurator.add_entry(_("Number of lines:"), "num_lines", _("The number of lines of history to display."))
        configurator.add_entry(_("Wrap after # characters:"), "max_chars", _("The max number of characters per line."))
        configurator.add_color_selector(_("Highlight color"), "highlight_color", _("The color to highlight messages with your nick in them."))
        configurator.add_option(_("Display message window with:"), "msg_window_display_action", _("Display the message window with a mouse click or on mouse over."), (("Mouse Click", "Mouse Click"), ("Mouse Over", "Mouse Over")))
        configurator.add_checkbox(_("Display time"), "bool_time", _("Display message timestamps."))
        configurator.add_title(_("Log Settings"))
        configurator.add_checkbox(_("Log chat"), "bool_log", _("Check to log chat to a file."))
        configurator.add_file_selector(_("Log file:"), "log_file", _("The log file."))
        configurator.add_title(_("Text to Speech Settings"))
        configurator.add_checkbox(_("Text to Speech"), "bool_tts", _("Check to invoke text to speech."))
        configurator.add_entry(_("TTS Command:"), "tts_command", _("The command to invoke the text to speech program."))
        return configurator

    def watch_borders(self, *args):
        if args[0][-1] == "width":
            self.update_border_width(args[-1])
        elif args[0][-1] == "height":
            self.update_border_height(args[-1])
        
    def update_border_width(self, value):
        output = self._new_output()
        width = int(value) - 29
        # Kludgy fix for problems when the display becomes
        # smaller. Because the dynamic borders would now
        # be larger than the desklet should be they cause
        # a catch-22... the desklets needs to be resized
        # smaller so that the borders will resize themselves
        # smaller, but the desklet can't be made smaller
        # because the borders are too big :(
        # So we are setting the value to zero and setting
        # a callback to redo it (giving it time to catch
        # back up with itself). This also affects the 
        # border height routine below
        if width < self.border_width:
            width = 0
            self._add_timer(1000, self.update_borders)
        self.border_width = width
        output.set("border_width", width)
        output.set("horizontal_divider_width", max(0, width - 70))
        self._send_output(output)
    
    def update_border_height(self, value):
        output = self._new_output()
        height = int(value) - 30
        if height < self.border_height:
            height = 0
            self._add_timer(1000, self.update_borders)
        output.set("border_height", height)
        self._send_output(output)

    def update_borders(self):
        geometry = self.get_desklet_location()
        width = int(geometry[2]) - 29
        height = int(geometry[3]) - 30
        output = self._new_output()
        output.set("border_width", width)
        output.set("border_height", height)
        output.set("horizontal_divider_width", max(0, width - 70))
        self._send_output(output)
        return 0

    def reconnect_desklet(self, widget, *args):
        self.ircbot_stop()
        self.ircbot_start()

    def _shutdown(self):
        self.ircbot_stop()
        try:    
            self.msg_window.destroy()
        except:
            pass

def new_sensor(args): return apply(IRCSensor, args)
