# IRC desklet for gDesklets (http://gdesklets.gnomedesktop.org)
# Copyright (C) 2003  Seth Remington

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from twisted.internet import reactor, protocol
from twisted.protocols import irc
import os
import time
import gtk
from main import admin


class IRCProtocol(irc.IRCClient):
    def lineReceived(self, data):
        #print "Received: " + data
        irc.IRCClient.lineReceived(self,
            data.decode(self.factory.encoding, "ignore").encode("utf-8"))

    def sendLine(self, line):
        #print "Sending: " + line
        irc.IRCClient.sendLine(self,
            line.decode("utf-8").encode(self.factory.encoding, "ignore"))

    def irc_RPL_YOURHOST(self, prefix, params):
        """ Overriding to capture hostname.
        This is as good a place as any :) Better way?"""
        self.yourHost(params[1])
        self.pinger = pinger(prefix, 30000, self)

    def parse_user(self, user):
        return user.split("!")[0]

    def escape(self, text):
        import xml.sax.saxutils
        return xml.sax.saxutils.escape(text)

    def register(self, nickname, hostname="foo", servername="bar"):
        # twisted.protocols.irc uses nickname not only for "nickname",
        # but also for "username" of USER command.  But, some letters
        # available on nickname may not allowed for username and it would
        # distrub connections.
        if self.password is not None:
            self.sendLine("PASS %s" % self.password)
        self.setNick(nickname)
        self.sendLine("USER %s foo bar :%s" % (self.username, self.realname))

    def connectionMade(self):
        self.username = self.factory.username
        self.nickname = self.factory.nickname
        self.password = self.factory.password
        irc.IRCClient.connectionMade(self)
        self.factory.protocol_instance = self
        
        self.logger = self.factory.logger
        self.log = self.logger.log
        self.displayer = self.factory.displayer
        self.display = self.displayer.log
        self.TTS = self.factory.TTS

    def signedOn(self):
        self.join(self.factory.channel)
        self.displayer.set_channel(self.factory.channel)

    def connectionLost(self, reason):
        irc.IRCClient.connectionLost(self)
        self.logger.close()

    def joined(self, channel):
        message = _("You have joined %s") % channel
        self.log(message)
        self.TTS.out(message)
        self.display(message)

    def privmsg(self, user, channel, message):
        user = self.parse_user(user)
        to_me = message.find(self.factory.nickname)
        self.log("<%s> %s" % (user, message))
        self.TTS.out(message)
        if(to_me != -1):
            to_me = 1
        else:
            to_me = 0
        self.display("&lt;%s&gt; %s" % (user, self.escape(message)), to_me)

    def action(self, user, channel, msg):
        user = self.parse_user(user)
        self.log("* %s %s" % (user, msg))
        self.TTS.out("%s %s" % (user, msg))
        self.display("* %s %s" % (user, self.escape(msg)))

    def userJoined(self, user, channel):
        user = self.parse_user(user)
        message = _("%(user)s has joined %(channel)s") % locals()
        self.log(message)
        self.TTS.out(message)
        self.display(message)
        self.displayer.add_client(user)

    def userLeft(self, user, channel):
        user = self.parse_user(user)
        message = _("%(user)s has left %(channel)s") % locals()
        self.log(message)
        self.TTS.out(message)
        self.display(message)
        self.displayer.remove_client(user)

    def irc_QUIT(self, prefix, params):
        nick = self.parse_user(prefix)
        message = _("%(nick)s has quit (%(reason)s)") % {
                    'nick': nick, 'reason': self.escape(params[0])}
        self.log(message)
        self.TTS.out(message)
        self.display(message)
        self.displayer.remove_client(nick)

    def irc_PING(self, prefix, params):
        """ We need to check if we've lost our connection
        to the server. We'll try listening for the pings
        and reconnect if we don't receive one in x seconds"""
        irc.IRCClient.irc_PING(self, prefix, params)

    def irc_RPL_NAMREPLY(self, prefix, params):
        # This is a horrible assumption since I'm sure there
        # will be an instance where it won't be true, but
        # until I can figure out a better way...
        for item in params[3:]:
            for client in item.decode("utf-8").split():
                self.displayer.add_client(client.encode("utf-8"))

    def irc_RPL_NOWAWAY(self, prefix, params):
        pass

    def irc_RPL_UNAWAY(self, prefix, params):
        pass

    def giveOps(self, user):
        self.sendLine("MODE %s +o %s" % (self.factory.channel, user))
    
    def giveVoice(self, user):
        self.sendLine("MODE %s +v %s" % (self.factory.channel, user))

    def takeOps(self, user):
        self.sendLine("MODE %s -o %s" % (self.factory.channel, user))
    
    def takeVoice(self, user):
        self.sendLine("MODE %s -v %s" % (self.factory.channel, user))

    def quit(self, message = "Quit"):
        self.sendLine("QUIT :%s" % (message))

    def userKicked(self, kickee, channel, kicker, message):
        message = _("%(kicker)s has kicked %(kickee)s from channel %(channel)s") % locals()
        self.log(message)
        self.TTS.out(message)
        self.display(message)
        self.displayer.remove_client(kickee)

    def userRenamed(self, oldname, newname):
        message = _("%(oldname)s is now known as %(newname)s") % locals()
        self.log(message)
        self.TTS.out(message)
        self.display(message)
        self.displayer.remove_client(oldname)
        self.displayer.add_client(newname)

    def nickChanged(self, nick):
        message = _("You are now known as %s") % (nick)
        self.log(message)
        self.TTS.out(message)
        self.display(message)
        self.displayer.remove_client(self.factory.nickname)
        self.displayer.add_client(nick)
        self.factory.change_nickname(nick)

    def topicUpdated(self, user, channel, newTopic):
        escapedTopic = self.escape(newTopic)
        message = _("%(user)s has set the topic for %(channel)s to %(escapedTopic)s") % locals()
        self.log(message)
        self.TTS.out(message)
        self.display(message)
        self.displayer.set_topic(self.factory.channel, escapedTopic)

    def modeChanged(self, user, channel, set, modes, args):
        if not args:
            return
        giver = self.parse_user(user)
        if set:
            give_take = "gives"
            verb = "to"
        else:
            give_take = "takes"
            verb = "from"
        if modes == "o":
            mode = "channel operator status"
        elif modes == "v":
            mode = "voice"
        receiver = args[0]
        if set and modes == "o":
            new_client = "@" + receiver
        elif set and modes == "v":
            new_client = "+" + receiver
        else:
            new_client = receiver

        message = "%s %s %s %s %s" % (giver, give_take, mode, verb, receiver)
        self.log(message)
        self.TTS.out(message)
        self.display(message)
        self.displayer.remove_client(receiver)
        self.displayer.add_client(new_client)

    def luserClient(self, info):
        print info

    def send_message(self, channel, message):
        if message[0] == "/":
            if message[1:7].lower() == "action":
                self.me(channel, message[7:].strip())
                self.action(self.factory.nickname, channel, message[7:].strip())
            elif message[1:3].lower() == "me":
                self.me(channel, message[3:].strip())
                self.action(self.factory.nickname, channel, message[3:].strip())
            elif message[1:5].lower() == "nick":
                self.setNick(message[5:].strip())
            elif message[1:6].lower() == "topic":
                self.topic(channel, message[6:].strip())
            elif message[1:5].lower() == "away":
                self.away(message[5:].strip())
            elif message[1:3].lower() == "op":
                self.giveOps(message[3:].strip())
            elif message[1:5].lower() == "deop":
                self.takeOps(message[5:].strip())
            elif message[1:6].lower() == "voice":
                self.giveVoice(message[6:].strip())
            elif message[1:8].lower() == "devoice":
                self.takeVoice(message[8:].strip())
            else:
                # Command not recognized
                pass
        else:
            self.msg(channel, message)
            # Server doesn't appear to send a copy to the client that sent the message
            # so we will send it to ourselves
            self.privmsg(self.factory.nickname, channel, message)

    def tab_complete(self, widget, start):
        if not start:
            return
        if widget.get_text().find(" ") == -1:
            matches = []
            # Check if we are completing a nick or an action
            if start[0] == "/":
                actions = ["ACTION", "ME", "NICK", "TOPIC", "AWAY", "OP", "VOICE", "DEOP", "DEVOICE"]
                for action in actions:
                    if start[1:].upper() == action[:len(start) - 1]:
                        matches.append(action)
                if len(matches):
                    if len(matches) == 1:
                        widget.set_text("/" + matches[0] + " ")
                        widget.set_position(-1)
                    elif len(matches) > 1:
                        self.display("--- " + " ".join(matches) + " ---")
            else:
                self.complete_nick(widget, start, ": ")
        else:
            self.complete_nick(widget, start)

    def complete_nick(self, widget, start, append = ""):
        matches = []
        for nick in self.displayer.clients:
            if start == nick[1][:len(start)]:
                matches.append(nick[1])
        if len(matches):
            if len(matches) == 1:
                text = widget.get_text()
                widget.set_text(text[:-(len(start))] + matches[0] + append)
                widget.set_position(-1)
            elif len(matches) > 1:
                self.display("--- " + " ".join(matches) + " ---")

    def irc_PONG(self, prefix, params):
        self.pinger.got_pong(time.time())

class IRCFactory(protocol.ClientFactory):
    protocol = IRCProtocol
    protocol_instance = None

    def __init__(self, sensor):
        self.sensor = sensor
        self.channel = self.sensor.channel
        self.nickname = self.sensor.nickname
        self.username = self.sensor.username
        self.password = self.sensor.password
        self.encoding = self.sensor.encoding
        self.logger = IRCLogger(self.sensor)
        self.displayer = IRCDisplay(self)
        self.TTS = TTS(self.sensor)
        global _; _ = self.sensor.gettext

    def clientConnectionLost(self, connector, reason):
        message = _("Lost connection to host '%s'") % (self.sensor.server)
        self.logger.log(message)
        self.TTS.out(message)
        self.displayer.log(message)
        if connector.irc_is_running:
            # We need to zero out the clients
            self.displayer.reset_clients()
            connector.connect()

    def clientConnectionFailed(self, connector, reason):
        output = self.sensor._new_output()
        output.set("msg", _("Could not connect to host '%s'") % (self.sensor.server))
        self.sensor._send_output(output)
        reactor.stop()

    def change_nickname(self, nick):
        self.nickname = nick

    def change_channel(self, channel):
        self.channel = channel

class IRCLogger:
    def __init__(self, sensor):
        self.sensor = sensor
        self.logfile = self.sensor.log_file
        if self.sensor.bool_log:
            self.file = open(self.logfile, "a")

    def log(self, message):
        if self.sensor.bool_log:
            self.file.write("%s\n" % message)
            self.file.flush()

    def close(self):
        if self.sensor.bool_log:
            self.file.close()

class IRCDisplay:
    def __init__(self, factory):
        self.factory = factory
        self.sensor = self.factory.sensor
        self.history = []
        self.clients = []
        self.scroll_hist = 0
        self.scroll_client = 0

    def scroll(self, direction):
        num = int(self.sensor.num_lines)
        def update():
            output = self.build_output()
            self.sensor_output("msg", output)
        # direction: up is 0, down is 1
        if not direction: # up
            if len(self.history) <= num: pass
            elif len(self.history) <= num + self.scroll_hist: pass
            else: self.scroll_hist += 1; update()
        else: # down
            if len(self.history) < num: pass
            elif self.scroll_hist > 0:
                self.scroll_hist -= 1
                update()

    def scroll_clients(self, direction):
        num = int(self.sensor.num_lines)
        def update():
            output = self.build_clients_output()
            self.sensor_output("clients", output)
        # direction: up is 0, down is 1
        if direction: # down
            if len(self.clients) <= num: pass
            elif len(self.clients) <= num + self.scroll_client: pass
            else: self.scroll_client += 1; update()
        else: # up
            if len(self.clients) < num: pass
            elif self.scroll_client > 0:
                self.scroll_client -= 1
                update()

    def add_client(self, client):
        if client[0] == "@": # Operator
            self.clients.append(["@", client[1:]])
        elif client[0] == "+": # Voice
            self.clients.append(["+", client[1:]])
        else:
            self.clients.append([None, client])
        #print self.clients
        output = self.build_clients_output()
        self.sensor_output("clients", output)
        
    def build_clients_output(self):
        output = []
        self.clients.sort(self.sort_clients)
        num = int(self.sensor.num_lines)
        if len(self.clients) < num: num = len(self.clients)
        for i in xrange(num):
            if self.clients[self.scroll_client + i][0]:
                tmp = self.clients[self.scroll_client + i][0] + " " + self.clients[self.scroll_client + i][1]
            else:
                tmp = self.clients[self.scroll_client + i][1]
            output.append(tmp)
        return "\n".join(output)

    def sort_clients(self, x, y):
        if x[0] == "@" and (y[0] == "+" or not y[0]):
            return -1
        if x[0] == "+" and y[0] == "@":
            return 1
        if x[0] == "+" and not y[0]:
            return -1
        if not x[0] and (y[0] == "+" or y[0] == "@"):
            return 1
        if x[0] == y[0]:
            return cmp(x[1].lower(), y[1].lower())

    def remove_client(self, client):
        num = int(self.sensor.num_lines)
        for c in self.clients:
            if c[1] == client:
                if len(self.clients) + self.scroll_client >= num and self.scroll_client:
                    self.scroll_client -= 1
                self.clients.remove([c[0], c[1]])
                output = self.build_clients_output()
                self.sensor_output("clients", output)
                break

    def reset_clients(self):
        self.clients = []
        
    def log(self, message, flags = 0):
        """ Logs message to display history to be displayed.
        
        Param "flags" can be set to 1 which means to highlight
        the line."""
        self.history.append((time.localtime(), message, flags))
        output = self.build_output()
        self.sensor_output("msg", output)

    def build_output(self):
        # Send the number of lines specified
        # in the configuration to the display.
        # Also display the proper slice of
        # history if the user has scrolled
        output = []
        num = int(self.sensor.num_lines)
        if len(self.history) < num: num = len(self.history)
        for i in xrange(num):
            if self.sensor.bool_time:
                line = time.strftime("[%H:%M:%S] ", self.history[-(num + self.scroll_hist) + i][0]) + self.history[-(num + self.scroll_hist) + i][1]
            else:
                line = self.history[-(num + self.scroll_hist) + i][1]
            line = self.wrap_output(line, int(self.sensor.max_chars), self.history[-(num + self.scroll_hist) + i][2])
            output.append(line)
        return "\n".join(output)

    def highlight_line(self, line):
            return "<span foreground='" + self.sensor.highlight_color + "'>" + line + "</span>"

    def wrap_output(self, line, length, highlight = 0):
        """ Split on word boundries where the total length
        of one line does not exceed length parameter.
        Should be passed only one line at a time."""
        wrapped_output = []
        line = line.decode("utf-8")
        if len(line) < length:
            if(highlight == 1):
                line = self.highlight_line(line.encode("utf-8"))
            wrapped_output.append(line)
        else:
            new_line = u""
            words = line.split()
            for word in words:
                # Just in case the word is longer than the max length
                if len(word) > length:
                    if new_line:
                        if(highlight == 1):
                            new_line = self.highlight_line(new_line)
                        wrapped_output.append(new_line)
                    if(highlight == 1):
                        word = self.highlight_line(word)
                    wrapped_output.append(word)
                    new_line = u""
                elif len(new_line) + len(word) + 1 > length:
                    if(highlight == 1):
                        new_line = self.highlight_line(new_line)
                    wrapped_output.append(new_line)
                    new_line = u"            " + word
                else:
                    if new_line:
                        new_line += u" " + word
                    else:
                        new_line = word
            if new_line:
                if(highlight == 1):
                    new_line = self.highlight_line(new_line)
                wrapped_output.append(new_line)
        output = u"\n".join(wrapped_output)
        return output.encode("utf-8")

    def set_channel(self, channel):
        self.sensor_output("channel", channel)

    def set_topic(self, channel, topic):
        self.sensor_output("topic",
            self.wrap_output(topic, int(self.sensor.max_chars)))

    def sensor_output(self, name, value):
        output = self.sensor._new_output()
        output.set(name, value)
        self.sensor._send_output(output)

class TTS:
    def __init__(self, sensor):
        self.sensor = sensor

    def out(self, message):
        if self.sensor.bool_tts:
            fi, fo = os.popen2(self.sensor.tts_command)
            fi.writelines(message)
            fi.close()

class pinger:
    def __init__(self, host, interval, irc_instance):
        self.last_ping = time.time()
        self.host = host
        self.interval = interval
        self.irc_instance = irc_instance
        gtk.timeout_add(self.interval, self.pinger_callback)

    def pinger_callback(self):
        if self.last_ping <= time.time() - 300:
            self.irc_instance.transport.loseConnection()
        self.irc_instance.sendLine("PING %s" % self.host)
        return gtk.TRUE

    def got_pong(self, time):
        self.last_ping = time
